/*
 * FixScript v0.8 - https://www.fixscript.org/
 * Copyright (c) 2018-2022 Martin Dvorak <jezek2@advel.cz>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <locale.h>
#include <pthread.h>
#include <unistd.h>

#define IMPORT(name) __attribute__ ((import_name(name))) 
#define EXPORT(name) __attribute__ ((export_name(name))) 

extern void print(const char *s) IMPORT("print");
extern void double_to_string(const char *buf, int size, double value, int num_decimals) IMPORT("double_to_string");
extern double string_to_double(const char *ptr) IMPORT("string_to_double");
extern int grow_memory(int num_pages) IMPORT("grow_memory");
extern double dabs(double x) IMPORT("fabs");
extern double dmin(double x, double y) IMPORT("fmin");
extern double dmax(double x, double y) IMPORT("fmax");
extern double dfloor(double x) IMPORT("floor");
extern double dceil(double x) IMPORT("ceil");
extern double dround(double x) IMPORT("round");
extern double dpow(double x, double y) IMPORT("pow");
extern double dsqrt(double x) IMPORT("sqrt");
extern double dcbrt(double x) IMPORT("cbrt");
extern double dexp(double x) IMPORT("exp");
extern double dlog(double x) IMPORT("log");
extern double dlog2(double x) IMPORT("log2");
extern double dlog10(double x) IMPORT("log10");
extern double dsin(double x) IMPORT("sin");
extern double dcos(double x) IMPORT("cos");
extern double dasin(double x) IMPORT("asin");
extern double dacos(double x) IMPORT("acos");
extern double dtan(double x) IMPORT("tan");
extern double datan(double x) IMPORT("atan");
extern double datan2(double y, double x) IMPORT("atan2");
extern double get_monotonic_time() IMPORT("get_monotonic_time");

static void *free_blocks[32];
static int heap_start, heap_end, heap_limit, heap_free;

extern int __heap_base;

int errno;


void *malloc(size_t size) EXPORT("malloc")
{
   void *block;
   int start, grow;
   
   if (size > 128*1024*1024) return NULL;
   if (size < 4) size = 4;
   size = 32 - __builtin_clz(size-1);

   block = free_blocks[size];
   if (block) {
      free_blocks[size] = *((void **)block);
      heap_free -= 4+(1<<size);
      return block;
   }

   if (heap_start == 0) {
      heap_start = (((intptr_t)&__heap_base) + 3) & ~3;
      heap_end = heap_start;
      heap_limit = __builtin_wasm_memory_size(0) << 16;
   }

   start = heap_end+4;
   if (size > 2) {
      start = (start+7)&~7;
   }

   if (start + (1<<size) > heap_limit) {
      grow = (start + (1<<size) - heap_limit + 0xFFFF) >> 16;
      if (grow < 16) grow = 16;
      if (grow_memory(grow) < 0) {
         return NULL;
      }
      heap_limit += grow << 16;
   }

   ((int *)start)[-1] = size;
   heap_end = start + (1<<size);
   return (void *)start;
}


void free(void *ptr) EXPORT("free")
{
   int size;
   
   if (!ptr) return;

   size = ((int *)ptr)[-1];
   *((void **)ptr) = free_blocks[size];
   free_blocks[size] = ptr;
   heap_free += 4+(1<<size);
}


void *calloc(size_t nmemb, size_t size) EXPORT("calloc")
{
   void *ptr;
   int64_t full_size = (int64_t)nmemb * (int64_t)size;

   if (full_size < 0 || full_size > 0x7FFFFFFF) {
      return NULL;
   }

   ptr = malloc(full_size);
   if (!ptr) {
      return NULL;
   }

   memset(ptr, 0, full_size);
   return ptr;
}


void *realloc(void *ptr, size_t size) EXPORT("realloc")
{
   void *new_ptr;
   int cur_size;

   if (!ptr) {
      return malloc(size);
   }

   cur_size = 1 << (((int *)ptr)[-1]);
   if (size <= cur_size) {
      return ptr;
   }

   new_ptr = malloc(size);
   if (!new_ptr) return NULL;

   memcpy(new_ptr, ptr, cur_size);
   free(ptr);
   return new_ptr;
}


int malloc_get_heap_used()
{
   return (heap_end - heap_start) - heap_free;
}


int malloc_get_heap_free()
{
   return heap_free;
}


int malloc_get_heap_size()
{
   return heap_end - heap_start;
}


void *memset(void *s, int c, size_t n)
{
   char *sc = s;
   uint64_t *start, *end, c8;

   if (n >= 16) {
      start = (uint64_t *)((((intptr_t)sc) + 7) & ~7);
      end = (uint64_t *)(((intptr_t)sc + n) & ~7);
      c8 = ((uint8_t)c) * 0x0101010101010101ULL;
      while (sc < (char *)start) {
         *sc++ = c;
      }
      while (start < end) {
         *start++ = c8;
      }
      sc = (char *)start;
      while (sc < ((char *)s)+n) {
         *sc++ = c;
      }
      return s;
   }

   while (n-- > 0) {
      *sc++ = c;
   }
   return s;
}


void *memcpy(void *dest, const void *src, size_t n)
{
   char *cdest = dest;
   const char *csrc = src;
   while (n-- > 0) {
      *cdest++ = *csrc++;
   }
   return dest;
}


void *memmove(void *dest, const void *src, size_t n)
{
   char *cdest = dest;
   const char *csrc = src;
   
   if ((intptr_t)cdest < (intptr_t)csrc) {
      while (n-- > 0) {
         *cdest++ = *csrc++;
      }
   }
   else {
      if (n == 0) return dest;
      cdest += n-1;
      csrc += n-1;
      while (n-- > 0) {
         *cdest-- = *csrc--;
      }
   }
   return dest;
}


int memcmp(const void *s1, const void *s2, size_t n)
{
   const char *c1 = s1;
   const char *c2 = s2;
   char v1, v2;
   while (n-- > 0) {
      v1 = *c1++;
      v2 = *c2++;
      if (v1 < v2) return -1;
      if (v1 > v2) return +1;
   }
   return 0;
}


char *strcpy(char *dest, const char *src)
{
   char *orig_dest = dest;
   while ((*dest++ = *src++));
   return orig_dest;
}


char *strcat(char *dest, const char *src)
{
   char *orig_dest = dest;
   for (; *dest; dest++);
   while ((*dest++ = *src++));
   return orig_dest;
}


size_t strlen(const char *s)
{
   size_t len = 0;
   while (*s++) {
      len++;
   }
   return len;
}


int strcmp(const char *s1, const char *s2)
{
   char c1, c2;
   for (;;) {
      c1 = *s1++;
      c2 = *s2++;
      if (c1 < c2) return -1;
      if (c1 > c2) return +1;
      if (c1 == 0) return 0;
   }
   return 0;
}


int strncmp(const char *s1, const char *s2, size_t n)
{
   char c1, c2;
   while (n-- > 0) {
      c1 = *s1++;
      c2 = *s2++;
      if (c1 < c2) return -1;
      if (c1 > c2) return +1;
      if (c1 == 0) return 0;
   }
   return 0;
}


char *strdup(const char *s)
{
   char *dup;
   int len;

   len = strlen(s);
   dup = malloc(len+1);
   if (!dup) return NULL;
   memcpy(dup, s, len+1);
   return dup;
}


char *strchr(const char *s, int c)
{
   for (; *s; s++) {
      if (*s == c) return (char *)s;
   }
   return NULL;
}


char *strrchr(const char *s, int c)
{
   const char *e = s;
   for (; *e; e++);
   for (; e >= s; e--) {
      if (*e == c) return (char *)e;
   }
   return NULL;
}


char *strstr(const char *haystack, const char *needle)
{
   const char *s = haystack, *s1, *s2;
   char c = *needle++, c1, c2;
   if (c == 0) return (char *)s;
   for (; *s; s++) {
      if (*s == c) {
         s1 = s+1;
         s2 = needle;
         for (;;) {
            c1 = *s1++;
            c2 = *s2++;
            if (c2 == 0) return (char *)s;
            if (c1 == 0) break;
            if (c1 != c2) break;
         }
      }
   }
   return NULL;
}


double strtod_l(const char *nptr, char **endptr, locale_t locale)
{
   double value;

   value = string_to_double(nptr);
   if (endptr) {
      *endptr = value == value? (char *)nptr + strlen(nptr) : (char *)nptr;
   }
   return value;
}


long int strtol(const char *nptr, char **endptr, int base)
{
   int value = 0, neg = 0;
   char c;
   if (strncmp(nptr, "-2147483648", 11) == 0) {
      if (endptr) *endptr = (char *)nptr+11;
      return 0x80000000;
   }
   if (*nptr == '+') {
      nptr++;
   }
   else if (*nptr == '-') {
      nptr++;
      neg = 1;
   }
   for (;;) {
      c = *nptr;
      if (c < '0' || c > '9') break;
      if (value > 214748364) {
         if (endptr) *endptr = (char *)nptr;
         return value;
      }
      value *= 10;
      c -= '0';
      if (c > 2147483647 - value) {
         if (endptr) *endptr = (char *)nptr;
         return value;
      }
      value += c;
      nptr++;
   }
   if (neg) value = -value;
   if (endptr) *endptr = (char *)nptr;
   return value;
}


long long int strtoll(const char *nptr, char **endptr, int base)
{
   int64_t value = 0, neg = 0;
   char c;
   if (strncmp(nptr, "-9223372036854775808", 20) == 0) {
      if (endptr) *endptr = (char *)nptr+20;
      return 0x8000000000000000ULL;
   }
   if (*nptr == '+') {
      nptr++;
   }
   else if (*nptr == '-') {
      nptr++;
      neg = 1;
   }
   for (;;) {
      c = *nptr;
      if (c < '0' || c > '9') break;
      if (value > 922337203685477580LL) {
         if (endptr) *endptr = (char *)nptr;
         return value;
      }
      value *= 10;
      c -= '0';
      if (c > 9223372036854775807LL - value) {
         if (endptr) *endptr = (char *)nptr;
         return value;
      }
      value += c;
      nptr++;
   }
   if (neg) value = -value;
   if (endptr) *endptr = (char *)nptr;
   return value;
}


unsigned long int strtoul(const char *nptr, char **endptr, int base)
{
   uint32_t value = 0;
   const char *start;
   char c;
   if (nptr[0] == '0' && nptr[1] == 'x') {
      nptr += 2;
   }
   start = nptr;
   for (;;) {
      c = *nptr;
      if (c >= '0' && c <= '9') {
         value = (value << 4) + (c - '0');
         nptr++;
         continue;
      }
      if (c >= 'a' && c <= 'f') {
         value = (value << 4) + (c - 'a' + 10);
         nptr++;
         continue;
      }
      if (c >= 'A' && c <= 'F') {
         value = (value << 4) + (c - 'A' + 10);
         nptr++;
         continue;
      }
      break;
   }
   if (endptr) {
      if (nptr - start > 8) {
         *endptr = (char *)start + 8;
      }
      else {
         *endptr = (char *)nptr;
      }
   }
   return value;
}


int atoi(const char *nptr)
{
   return strtol(nptr, NULL, 10);
}


void qsort(void *base, size_t nmemb, size_t size, int(*compar)(const void *, const void *))
{
   int i, left = 0, right = ((int)nmemb)-1, left_idx = left, right_idx = right, pivot;
   char *p1, *p2, tmp;

   if (right_idx > left_idx) {
      pivot = left + ((right - left) >> 1);
      while (left_idx <= pivot && right_idx >= pivot) {
         for (;;) {
            if (compar(base + left_idx*size, base + pivot*size) < 0 && left_idx <= pivot) {
               left_idx++;
               continue;
            }
            break;
         }
         for (;;) {
            if (compar(base + pivot*size, base + right_idx*size) < 0 && right_idx >= pivot) {
               right_idx--;
               continue;
            }
            break;
         }
         p1 = base + left_idx*size;
         p2 = base + right_idx*size;
         for (i=0; i<size; i++) {
            tmp = *p1;
            *p1++ = *p2;
            *p2++ = tmp;
         }
         left_idx++;
         right_idx--;
         if (left_idx - 1 == pivot) {
            pivot = right_idx = right_idx + 1;
         }
         else if (right_idx + 1 == pivot) {
            pivot = left_idx = left_idx - 1;
         }
      }
      if (pivot > 0) qsort(base + left*size, (size_t)(pivot - left), size, compar);
      qsort(base + (pivot+1)*size, (size_t)(right - (pivot + 1) + 1), size, compar);
   }
}


int abs(int j)
{
   return j < 0? -j : j;
}


int fprintf(FILE *stream, const char *format, ...)
{
   char buf[4096];
   va_list ap;

   va_start(ap, format);
   vsnprintf(buf, sizeof(buf), format, ap);
   va_end(ap);
   print(buf);
   return 0;
}


int sprintf(char *str, const char *format, ...)
{
   va_list ap;
   int ret;

   va_start(ap, format);
   ret = vsnprintf(str, 0x7FFFFFFF, format, ap);
   va_end(ap);
   return ret;
}


int snprintf(char *str, size_t size, const char *format, ...)
{
   va_list ap;
   int ret;

   va_start(ap, format);
   ret = vsnprintf(str, size, format, ap);
   va_end(ap);
   return ret;
}


static void int_to_string(char *str, int value)
{
   char buf[12], *s = buf+12;
   int sign = 0;

   if (value == 0x80000000) {
      strcpy(str, "-2147483648");
      return;
   }
   if (value == 0) {
      strcpy(str, "0");
      return;
   }

   if (value < 0) {
      value = -value;
      sign = 1;
   }

   *(--s) = 0;
   while (value) {
      *(--s) = (value % 10) + '0';
      value /= 10;
   }
   if (sign) {
      *(--s) = '-';
   }
   strcpy(str, s);
}


static void long_to_string(char *str, int64_t value)
{
   char buf[21], *s = buf+21;
   int sign = 0;

   if (value == 0x8000000000000000ULL) {
      strcpy(str, "-9223372036854775808");
      return;
   }
   if (value == 0) {
      strcpy(str, "0");
      return;
   }

   if (value < 0) {
      value = -value;
      sign = 1;
   }

   *(--s) = 0;
   while (value) {
      *(--s) = (value % 10) + '0';
      value /= 10;
   }
   if (sign) {
      *(--s) = '-';
   }
   strcpy(str, s);
}


int vsnprintf(char *str, size_t size, const char *format, va_list ap)
{
   int i, cnt=0, limit, num_digits;
   char c, *s2, buf[64];

   if (size > 0x7FFFFFFF) {
      size = 0x7FFFFFFF;
   }
   limit = str? size : -1;

   for (;;) {
      c = *format++;
      if (!c) break;

      if (c == '%') {
         c = *format++;
         switch (c) {
            case '%':
               if (cnt++ < limit) {
                  *str++ = c;
               }
               break;

            case 's':
               s2 = va_arg(ap, char *);
               for (;;) {
                  c = *s2++;
                  if (!c) break;
                  if (cnt++ < limit) {
                     *str++ = c;
                  }
               }
               break;

            case 'd':
            case 'p':
               int_to_string(buf, va_arg(ap, int));
               s2 = buf;
               for (;;) {
                  c = *s2++;
                  if (!c) break;
                  if (cnt++ < limit) {
                     *str++ = c;
                  }
               }
               break;

            case 'l':
               if (*format++ == 'l') {
                  if (*format++ == 'd') {
                     long_to_string(buf, va_arg(ap, int64_t));
                     s2 = buf;
                     for (;;) {
                        c = *s2++;
                        if (!c) break;
                        if (cnt++ < limit) {
                           *str++ = c;
                        }
                     }
                  }
               }
               break;

            case 'f':
               double_to_string(buf, sizeof(buf), va_arg(ap, double), -1);
               s2 = buf;
               for (;;) {
                  c = *s2++;
                  if (!c) break;
                  if (cnt++ < limit) {
                     *str++ = c;
                  }
               }
               break;

            case '.':
               num_digits = 0;
               for (;;) {
                  c = *format++;
                  if (c >= '0' && c <= '9') {
                     num_digits = num_digits*10 + (c - '0');
                  }
                  else break;
               }
               if (c == '*') {
                  num_digits = va_arg(ap, int);
                  c = *format++;
               }
               switch (c) {
                  case 's':
                     s2 = va_arg(ap, char *);
                     for (i=0; i<num_digits; i++) {
                        if (cnt++ < limit) {
                           *str++ = *s2++;
                        }
                     }
                     break;

                  case 'f':
                  case 'g':
                     double_to_string(buf, sizeof(buf), va_arg(ap, double), num_digits);
                     s2 = buf;
                     for (;;) {
                        c = *s2++;
                        if (!c) break;
                        if (cnt++ < limit) {
                           *str++ = c;
                        }
                     }
                     break;
               }
               break;
         }
      }
      else {
         if (cnt++ < limit) {
            *str++ = c;
         }
      }
   }
   if (limit >= 0) {
      *str = 0;
   }
   return cnt;
}


FILE *fopen(const char *path, const char *mode)
{
   return NULL;
}


size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
   return 0;
}


size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
   fprintf(stderr, "%.*s", (int)size*(int)nmemb, (const char *)ptr);
   return 0;
}


int fflush(FILE *stream)
{
   return 0;
}


int ferror(FILE *stream)
{
   return 0;
}


int fclose(FILE *fp)
{
   return 0;
}


locale_t newlocale(int category, const char *name, locale_t base)
{
   return (void *)1;
}


void freelocale(locale_t locale)
{
}


int clock_gettime(clockid_t clk_id, struct timespec *tp)
{
   double time = get_monotonic_time();
   tp->tv_sec = time / 1000.0;
   tp->tv_nsec = (time - tp->tv_sec * 1000.0) * 1000000.0;
   return 0;
}


float fabsf(float x)
{
   return dabs(x);
}


float fminf(float x, float y)
{
   return dmin(x, y);
}


float fmaxf(float x, float y)
{
   return dmax(x, y);
}


float floorf(float x)
{
   return dfloor(x);
}


float ceilf(float x)
{
   return dceil(x);
}


float roundf(float x)
{
   return dround(x);
}


float powf(float x, float y)
{
   return dpow(x, y);
}


float sqrtf(float x)
{
   return dsqrt(x);
}


float cbrtf(float x)
{
   return dcbrt(x);
}


float expf(float x)
{
   return dexp(x);
}


float logf(float x)
{
   return dlog(x);
}


float log2f(float x)
{
   return dlog2(x);
}


float log10f(float x)
{
   return dlog10(x);
}


float sinf(float x)
{
   return dsin(x);
}


float cosf(float x)
{
   return dcos(x);
}


float asinf(float x)
{
   return dasin(x);
}


float acosf(float x)
{
   return dacos(x);
}


float tanf(float x)
{
   return dtan(x);
}


float atanf(float x)
{
   return datan(x);
}


float atan2f(float y, float x)
{
   return datan2(y, x);
}


double fabs(double x)
{
   return dabs(x);
}


double fmin(double x, double y)
{
   return dmin(x, y);
}


double fmax(double x, double y)
{
   return dmax(x, y);
}


double floor(double x)
{
   return dfloor(x);
}


double ceil(double x)
{
   return dceil(x);
}


double round(double x)
{
   return dround(x);
}


double pow(double x, double y)
{
   return dpow(x, y);
}


double sqrt(double x)
{
   return dsqrt(x);
}


double cbrt(double x)
{
   return dcbrt(x);
}


double exp(double x)
{
   return dexp(x);
}


double log(double x)
{
   return dlog(x);
}


double log2(double x)
{
   return dlog2(x);
}


double log10(double x)
{
   return dlog10(x);
}


double sin(double x)
{
   return dsin(x);
}


double cos(double x)
{
   return dcos(x);
}


double asin(double x)
{
   return dasin(x);
}


double acos(double x)
{
   return dacos(x);
}


double tan(double x)
{
   return dtan(x);
}


double atan(double x)
{
   return datan(x);
}


double atan2(double y, double x)
{
   return datan2(y, x);
}


int pthread_mutex_init(pthread_mutex_t *mutex, pthread_mutexattr_t *attr)
{
   return 0;
}


int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
   return 0;
}


int pthread_mutex_lock(pthread_mutex_t *mutex)
{
   return 0;
}


int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
   return 0;
}


int pthread_cond_init(pthread_cond_t *cond, void *attr)
{
   return 0;
}


int pthread_cond_destroy(pthread_cond_t *cond)
{
   return 0;
}


int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
   return 0;
}


int pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *abstime)
{
   return 0;
}


int pthread_cond_signal(pthread_cond_t *cond)
{
   return 0;
}


int pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*start_routine)(void *), void *arg)
{
   start_routine(arg);
   return 0;
}


int pthread_detach(pthread_t thread)
{
   return 0;
}


int pthread_mutexattr_init(pthread_mutexattr_t *attr)
{
   return 0;
}


int pthread_mutexattr_settype(pthread_mutexattr_t *attr, int type)
{
   return 0;
}


long sysconf(int name)
{
   if (name == _SC_NPROCESSORS_ONLN) return 1;
   return -1;
}
